#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from glsl_math import *




class solid_angle_line_t:
    def __init__(self):
        self.begin = zeros((3,))
        self.line_direction = zeros((3,))
        self.rcp_min_distance = 0.0
        self.scaled_begin_s = 0.0
        self.begin_s = 0.0
        self.extent_s = 0.0
        self.rcp_density = 0.0


class projected_solid_angle_line_t:
    def __init__(self):
        self.solid = solid_angle_line_t()
        self.normal = zeros((2,))
        self.begin_cdf = 0.0
        self.extent_cdf = 0.0


def clip_line(plane_origin, plane_normal, line_begin, line_end, line_direction, line_length):
    normal_s = dot(plane_normal, line_direction)
    normal_dot_line_begin = dot(plane_normal, line_begin - plane_origin)
    intersection_s = -normal_dot_line_begin / normal_s
    begin_s = where((normal_dot_line_begin < 0.0) ,intersection_s ,0.0)
    end_s = where((fma(line_length, normal_s, normal_dot_line_begin) < 0.0) ,intersection_s ,line_length)
    line_end[0:3] = fma(line_direction, vec3(end_s), line_begin)
    line_begin[0:3] = fma(line_direction, vec3(begin_s), line_begin)
    line_length[0] = end_s - begin_s


def prepare_solid_angle_line_sampling(line_begin, line_direction, line_length):
    line = solid_angle_line_t()
    line.begin = line_begin
    line.line_direction = line_direction
    line.scaled_begin_s = dot(line_direction, line_begin)
    sq_begin_distance = dot(line_begin, line_begin)
    sq_min_distance = fma(-line.scaled_begin_s, line.scaled_begin_s, sq_begin_distance)
    line.rcp_min_distance = inversesqrt(sq_min_distance)
    line.begin_s = line.scaled_begin_s * inversesqrt(sq_begin_distance)
    scaled_end_s = line.scaled_begin_s + line_length
    rcp_end_distance = inversesqrt(fma(scaled_end_s, scaled_end_s, sq_min_distance))
    line.extent_s = fma(scaled_end_s, rcp_end_distance, -line.begin_s)
    line.rcp_density = 2.0 * line.rcp_min_distance * line.extent_s
    return line


def sample_solid_angle_line(line, random_number):
    s = fma(random_number, line.extent_s, line.begin_s)
    t = sqrt(maximum(0.0, fma(-s, s, 1.0)))
    y = line.rcp_min_distance * t
    x = fma(-y, line.scaled_begin_s, s)
    return x * line.line_direction + y * line.begin


def evaluate_line_sampling_cdf(normal, s, cdf_begin):
    s = clip(s, -1.0, 1.0)
    angle = arcsin(s)
    t = sqrt(fma(-s, s, 1.0))
    return fma(dot(normal, vec2(s, t)), s, fma(angle, normal[1], -cdf_begin))


def prepare_projected_solid_angle_line_sampling_without_clipping(surface_normal, line_begin, line_direction, line_length):
    line = projected_solid_angle_line_t()
    line.solid = prepare_solid_angle_line_sampling(line_begin, line_direction, line_length)
    line.normal[0] = dot(surface_normal, line_direction)
    line.normal[1] = fma(-line.solid.scaled_begin_s, line.normal[0], dot(surface_normal, line_begin)) * line.solid.rcp_min_distance
    line.begin_cdf = evaluate_line_sampling_cdf(line.normal, line.solid.begin_s, 0.0)
    line.extent_cdf = evaluate_line_sampling_cdf(line.normal, line.solid.begin_s + line.solid.extent_s, line.begin_cdf)
    line.solid.rcp_density = line.solid.rcp_min_distance * abs(line.extent_cdf)
    line.solid.rcp_density = where((line.solid.rcp_density > 0.0) ,line.solid.rcp_density ,0.0)
    return line


def prepare_projected_solid_angle_line_sampling(surface_normal, line_begin, line_direction, line_length):
    line = projected_solid_angle_line_t()
    line.solid.begin = line_begin
    line.solid.line_direction = line_direction
    line_end = zeros((3,))
    clip_line(vec3(0.0), surface_normal, line_begin, line_end, line_direction, line_length)
    if (line_length <= 0.0):
        line.solid.rcp_density = 0.0
        return line
    else:
        return prepare_projected_solid_angle_line_sampling_without_clipping(surface_normal, line_begin, line_direction, line_length)


def is_projected_solid_angle_line_empty(line):
    return line.solid.rcp_density <= 0.0


def solve_symmetric_homogeneous_quadratic(quadratic):
    coeff_xy = quadratic[1][0]
    sqrt_discriminant = sqrt(maximum(0.0, coeff_xy * coeff_xy - quadratic[0][0] * quadratic[1][1]))
    scaled_root = abs(coeff_xy) + sqrt_discriminant
    return  where((coeff_xy >= 0.0) ,vec2(scaled_root, -quadratic[0][0]) ,vec2(quadratic[1][1], scaled_root))


def invert_line_sampling_cdf(line, random_number, cdf):
    initial = zeros((2,))
    initial[0] = fma(random_number, line.solid.extent_s, line.solid.begin_s)
    initial[0] = clip(initial[0], -1.0, 1.0)
    initial[1] = sqrt(fma(-initial[0], initial[0], 1.0))
    dir = initial
    angle = arcsin(dir[0])
    for i in range(0, 2):
        dot_0 = dot(line.normal, dir)
        objective = fma(dot_0, dir[0], fma(angle, line.normal[1], -cdf))
        half_derivative_1 = dot_0 * dir[1]
        dot_1 = dot(line.normal, vec2(dir[1], -dir[0]))
        half_derivative_2 = dot_1 * dir[1] - dot_0 * dir[0]
        quadratic = mat2(half_derivative_2, half_derivative_1, half_derivative_1, objective)
        root = solve_symmetric_homogeneous_quadratic(-quadratic)
        angle += root[0] / root[1]
        angle = clip(angle, -0.5 * pi, 0.5 * pi)
        dir[0] = sin(angle)
        dir[1] = cos(angle)
    acceptable_error = 1.0e-5
    use_iterations = (abs(random_number - 0.5) <= 0.5 - acceptable_error)
    dir[0] = where(use_iterations ,dir[0] ,initial[0])
    dir[1] = where(use_iterations ,dir[1] ,initial[1])
    return dir


def sample_projected_solid_angle_line(line, random_number):
    cdf = fma(random_number, line.extent_cdf, line.begin_cdf)
    dir = invert_line_sampling_cdf(line, random_number, cdf)
    y = line.solid.rcp_min_distance * dir[1]
    x = fma(-y, line.solid.scaled_begin_s, dir[0])
    return x * line.solid.line_direction + y * line.solid.begin


def compute_projected_solid_angle_line_sampling_error(line, random_number, sampled_dir):
    cdf = fma(random_number, line.extent_cdf, line.begin_cdf)
    error = evaluate_line_sampling_cdf(line.normal, dot(line.solid.line_direction, sampled_dir), cdf)
    return error / line.extent_cdf
