#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from glsl_math import *
from cubic_solver_ported import solve_cubic
from line_sampling_ported import *




class linear_cosine_warp_line_hart_t:
    def __init__(self):
        self.line = solid_angle_line_t()
        self.density_0 = 0.0
        self.density_1 = 0.0


def prepare_linear_cosine_warp_line_sampling_hart(surface_normal, line_begin, line_direction, line_length):
    line = linear_cosine_warp_line_hart_t()
    line_end = zeros((3,))
    clip_line(vec3(0.0), surface_normal, line_begin, line_end, line_direction, line_length)
    if (line_length <= 0.0):
        line.line.rcp_density = 0.0
        return line
    line.line = prepare_solid_angle_line_sampling(line_begin, line_direction, line_length)
    line.density_0 = maximum(0.0, dot(line_begin, surface_normal) * inversesqrt(dot(line_begin, line_begin)))
    line.density_1 = maximum(0.0, dot(line_end, surface_normal) * inversesqrt(dot(line_end, line_end)))
    density_sum = line.density_0 + line.density_1
    normalization = 2.0 / (line.line.rcp_density * density_sum)
    line.density_0 *= normalization
    line.density_1 *= normalization
    return line


def mix_fma(x, y, a):
    return fma(a, y, fma(-a, x, x))


def linear_warp(random_number, density_0, density_1):
    lerped_density_sq = mix_fma(density_0 * density_0, density_1 * density_1, random_number)
    divisor = density_0 + sqrt(lerped_density_sq)
    return random_number * (density_0 + density_1) / divisor


def sample_linear_cosine_warp_line_hart(out_density, line, random_number):
    random_number = linear_warp(random_number, line.density_0, line.density_1)
    out_density = mix_fma(line.density_0, line.density_1, random_number)
    return sample_solid_angle_line(line.line, random_number)


class quadratic_cosine_warp_line_hart_t:
    def __init__(self):
        self.line = solid_angle_line_t()
        self.density_0 = 0.0
        self.density_1 = 0.0
        self.density_2 = 0.0


def prepare_quadratic_cosine_warp_line_sampling_hart(surface_normal, line_begin, line_direction, line_length):
    line = quadratic_cosine_warp_line_hart_t()
    line_end = zeros((3,))
    clip_line(vec3(0.0), surface_normal, line_begin, line_end, line_direction, line_length)
    if (line_length <= 0.0):
        line.line.rcp_density = 0.0
        return line
    line.line = prepare_solid_angle_line_sampling(line_begin, line_direction, line_length)
    line.density_0 = maximum(0.0, dot(line_begin, surface_normal) * inversesqrt(dot(line_begin, line_begin)))
    line.density_2 = maximum(0.0, dot(line_end, surface_normal) * inversesqrt(dot(line_end, line_end)))
    middle = sample_solid_angle_line(line.line, 0.5)
    line.density_1 = maximum(0.0, dot(middle, surface_normal))
    density_sum = line.density_0 + line.density_1 + line.density_2
    normalization = 3.0 / (line.line.rcp_density * density_sum)
    line.density_0 *= normalization
    line.density_1 *= normalization
    line.density_2 *= normalization
    return line


def quadratic_warp(random_number, density_0, density_1, density_2):
    quadratic = vec3(density_0, 2.0 * (density_1 - density_0), density_0 - 2.0 * density_1 + density_2)
    cubic = vec4(0.0, quadratic[0], 0.5 * quadratic[1], (1.0 / 3.0) * quadratic[2])
    random_number *= dot(cubic.yzw, vec3(1.0))
    cubic[0] = -random_number
    roots = zeros((3,))
    if (solve_cubic(roots, cubic)):
        result = roots[0]
        result = where((roots[1] >= 0.0 and roots[1] <= 1.0) ,roots[1] ,result)
        result = where((roots[2] >= 0.0 and roots[2] <= 1.0) ,roots[2] ,result)
        return result
    else:
        return roots[0]


def quadratic_bezier(b_0_0, b_0_1, b_0_2, location):
    b_1_0 = mix_fma(b_0_0, b_0_1, location)
    b_1_1 = mix_fma(b_0_1, b_0_2, location)
    return mix_fma(b_1_0, b_1_1, location)


def sample_quadratic_cosine_warp_line_hart(out_density, line, random_number):
    random_number = quadratic_warp(random_number, line.density_0, line.density_1, line.density_2)
    out_density = quadratic_bezier(line.density_0, line.density_1, line.density_2, random_number)
    return sample_solid_angle_line(line.line, random_number)


class projected_solid_angle_line_li_t:
    def __init__(self):
        self.begin = zeros((3,))
        self.line_direction = zeros((3,))
        self.closest_point = zeros((3,))
        self.normal = zeros((2,))
        self.min_distance = 0.0
        self.begin_l = 0.0
        self.end_l = 0.0
        self.cdf_begin = 0.0
        self.rcp_density = 0.0


def get_line_sampling_unnormalized_cdf_li(line, l):
    angle = arctan(l / line.min_distance)
    d = line.min_distance
    d2 = d * d
    s = l / (d * fma(l, l, d2))
    return (s + angle / d2) * line.normal[1] + l * s * line.normal[0]


def get_line_sampling_cdf_li(line, l):
    return (get_line_sampling_unnormalized_cdf_li(line, l) - line.cdf_begin) / line.rcp_density


def get_line_sampling_pdf_li(line, l):
    dist_sq = l * l + line.min_distance * line.min_distance
    return 2.0 * line.min_distance * (line.normal[1] + l * line.normal[0]) / (line.rcp_density * dist_sq * dist_sq)


def prepare_projected_solid_angle_line_sampling_li_without_clipping(surface_normal, line_begin, line_direction, line_length):
    line = projected_solid_angle_line_li_t()
    line.begin = line_begin
    line.line_direction = line_direction
    line.begin_l = dot(line.line_direction, line.begin)
    line.end_l = line.begin_l + line_length
    line.closest_point = fma(vec3(-line.begin_l), line.line_direction, line.begin)
    line.min_distance = length(line.closest_point)
    line.normal[0] = dot(surface_normal, line.line_direction)
    line.normal[1] = dot(surface_normal, line.closest_point)
    line.cdf_begin = get_line_sampling_unnormalized_cdf_li(line, line.begin_l)
    cdf_end = get_line_sampling_unnormalized_cdf_li(line, line.end_l)
    line.rcp_density = cdf_end - line.cdf_begin
    return line


def sample_projected_solid_angle_line_li(line, random_number):
    left = line.begin_l
    right = line.end_l
    current = 0.5 * (left + right)
    for i in range(0, 20):
        no_bisection = (current >= left and current <= right)
        current = where(no_bisection ,current ,(0.5 * (left + right)))
        error = get_line_sampling_cdf_li(line, current) - random_number
        if (abs(error) < 1.0e-5 or i == 19):
            break
        left = where((error > 0.0) ,left ,current)
        right = where((error > 0.0) ,current ,right)
        derivative = get_line_sampling_pdf_li(line, current)
        current -= error / derivative
    return normalize(fma(vec3(current), line.line_direction, line.closest_point))
